/*
 * Copyright (C) 2004, 2005 Michael Dominic K. (http://www.mdk.org.pl)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

using System;
using System.IO;
using System.Collections;
using MovieDatabase;
using Util;

public enum strhChunkType { auds, vids };

public class RIFFReaderException : ApplicationException
{
    public RIFFReaderException(string _Msg) : base(_Msg)
    {
    }

}


public class Chunk 
{
    public char[] FourCC;
    public UInt32 Size;
    public char[] FormType;
    public byte[] Data;
    public MemoryStream StreamToData;

    // Translatable Strings
    public static readonly string ___ChunksError = "The Chunk structure of the AVI header is corrupted! Can't continue...";
    public static readonly string ___InvalidStrh = "Unknown strh Chunk type! This is not a valid AVI file...";
    public static readonly string ___InvalidSize = "The size of the Chunk is invalid! AVI file corrupted.";

    public Chunk(Chunk _CH)
    {
        ReadFromStream(_CH.StreamToData);
    }

    public void ReadFromStream(Stream _Stream)
    {
        BinaryReader __BR = new BinaryReader(_Stream);
        FourCC = __BR.ReadChars(4);
        Size = __BR.ReadUInt32();

        if (FourCCToString(FourCC) == "LIST" || FourCCToString(FourCC) == "RIFF") {
            FormType = __BR.ReadChars(4);
            Size = Size - 4;
        }
        else FormType = null;
    }

    public Chunk(Stream _Stream)
    {
        ReadFromStream(_Stream);
    }

    public static string FourCCToString(char[] _FC)
    {
        string __S = String.Empty;
        foreach(char __C in _FC)
            __S += __C;

        return __S;
    }

    public void ReadData(Chunk _CH)
    {
        ReadData(_CH.StreamToData);
    }


    public void ReadData(Stream _Stream)
    {
        if (Size > 66000) throw new RIFFReaderException(___InvalidSize + "(" + Size + "," + FourCCToString(FourCC) + ")");
        BinaryReader __BR = new BinaryReader(_Stream);
        Data = __BR.ReadBytes((int) Size);
        StreamToData = new MemoryStream(Data, 0, (int) Size);
    }

    public void Parse(int _Lev, ArrayList _Array)
    {
        while(StreamToData.Position < StreamToData.Length) {
            Chunk __Chunk;
            try {
                __Chunk = new Chunk(this);
                __Chunk.ReadData(StreamToData);
                __Chunk.SpitDebug(_Lev);
                if (FourCCToString(__Chunk.FourCC) == "LIST") __Chunk.Parse(_Lev+1, _Array);
                else _Array.Add(__Chunk);
            } catch (Exception __Excp) {
                throw new RIFFReaderException(___ChunksError);
            }
        }
        
        string __Align = String.Empty;

        for(int __i = 0; __i < _Lev; __i++)
            __Align += "   ";
        Debug.Message("AVIGet.dll", "RIFFReader", String.Format(__Align + "EOF chunk"));
        
    }


    public strhChunkType StrhChunkType() 
    {
        BinaryReader __BR = new BinaryReader(StreamToData);
        char[] __strhtype = __BR.ReadChars(4);
        if (FourCCToString(__strhtype) == "auds") return strhChunkType.auds;
        else if (FourCCToString(__strhtype) == "vids") return strhChunkType.vids;
        else throw new RIFFReaderException(___InvalidStrh);
        
    }

    
    public void ParseAudsChunk(out UInt16 _Codec)
    {
        Debug.Message("AVIGet.dll", "RIFFReader", "Found audio stream information");
        BinaryReader __BR = new BinaryReader(StreamToData);
        _Codec = __BR.ReadUInt16(); // Audio codec
        Debug.Message("AVIGet.dll", "RIFFReader", String.Format("CODEC: {0} (MP3: {1} AC3: {2})" , _Codec , 0x55, 0x2000));
    }


    public void ParseVidsChunk(out UInt32 _Width, out UInt32 _Height, out char[] _Codec)
    {
        Debug.Message("AVIGet.dll", "RIFFReader", "Found video stream information");
        BinaryReader __BR = new BinaryReader(StreamToData);
        __BR.ReadChars(4); // 4 chars empty
        _Width = __BR.ReadUInt32(); // X resolution
        _Height = __BR.ReadUInt32(); // Y resolution
        __BR.ReadChars(4); // 4 chars empty
        _Codec = __BR.ReadChars(4); // Codec FOURCC
        Debug.Message("AVIGet.dll", "RIFFReader", String.Format("WIDTH: {0} HEIGHT: {1} CODEC: {2}" , _Width, _Height, FourCCToString(_Codec)));
        
    }



    
    public void SpitDebug(int _Lev)
    {
        string __Align = String.Empty;

        for(int __i = 0; __i < _Lev; __i++)
            __Align += "   ";
        
        if (FormType != null) Debug.Message("AVIGet.dll", "RIFFReader", String.Format(__Align + "CHUNK [FOURCC: {0} SIZE: {1} FORMTYPE: {2}]", FourCCToString(FourCC), Size, FourCCToString(FormType)));
        else Debug.Message("AVIGet.dll", "RIFFReader", String.Format(__Align + "CHUNK [FOURCC: {0} SIZE {1}]", FourCCToString(FourCC), Size));
    }

}



public class RIFFReader {

    FileStream FS;
    UInt32 AVIWidth = 0;
    UInt32 AVIHeight = 0 ;
    UInt16 AVIAudioCodec = 0;
    char[] AVIVideoCodec = null; 

    // Translatable strings
    public static readonly string ___InvalidAVI = "This is not a valid AVI file! RIFF header not found.";
    public static readonly string ___FileNotFound = "was not found on the filesystem.";

    public RIFFReader(string _Filename)
    {
        try {
            FS = new FileStream(_Filename, FileMode.Open, FileAccess.Read);
        } catch (Exception __Excp) {
            throw new RIFFReaderException(_Filename + " " +  ___FileNotFound);
        }
    
    }


    public void Run()
    {
        ArrayList __InterestingChunks = new ArrayList();
        
        // "RIFF"
        Chunk __RIFFChunk = new Chunk(FS);
        if (Chunk.FourCCToString(__RIFFChunk.FourCC) != "RIFF") throw new RIFFReaderException(___InvalidAVI);

        // "LIST"
        Chunk __LISTChunk = new Chunk(FS);
        if (Chunk.FourCCToString(__LISTChunk.FourCC) != "LIST") throw new RIFFReaderException(___InvalidAVI);

        // Let's start parsing...

        __LISTChunk.ReadData(FS);
        __LISTChunk.SpitDebug(0);
        __LISTChunk.Parse(1, __InterestingChunks);

        // Video first as default
        strhChunkType __CurrentStreamType = strhChunkType.vids;

        foreach (Chunk __CH in __InterestingChunks) {
            if (Chunk.FourCCToString(__CH.FourCC) == "strh") __CurrentStreamType = __CH.StrhChunkType();
            if (Chunk.FourCCToString(__CH.FourCC) == "strf") {
                if (__CurrentStreamType == strhChunkType.auds) __CH.ParseAudsChunk(out AVIAudioCodec);
                else if (__CurrentStreamType == strhChunkType.vids) __CH.ParseVidsChunk(out AVIWidth, out AVIHeight, out AVIVideoCodec);
            }
            
        }

    }

    
    /* Fill the movie with acquired data... */
    public Movie GetMovie()
    {
        Movie __M = new Movie();

        // Resolution
        if (AVIWidth != 0 && AVIHeight != 0) __M.Data[(int) MovieField.QUALITY] = AVIWidth + "x" + AVIHeight;

        // Video Codec
        string __VCodecName = String.Empty;
        switch (Chunk.FourCCToString(AVIVideoCodec)) {
        case "XVID":
            __VCodecName="XviD MPEG-4 Codec";
            break;
        case "DIV3":
            __VCodecName="DivX 3.11 Low Motion";
            break;
        case "DIV4":
            __VCodecName="DivX 3.11 High Motion";
            break;
        case "DIVX":
        case "DX50":
            __VCodecName="DivX 5.x";
            break;
        }
        __M.Data[(int) MovieField.VIDEOFORMAT] = __VCodecName;


        // Audio Codec
        string __ACodecName = String.Empty;
        switch (AVIAudioCodec) {
        case 0x55:
            __ACodecName="MPEG Layer 3 (MP3)";
            break;
        case 0x50:
            __ACodecName="MPEG Layer 1 or 2 (MP1/MP2)";
            break;
        case 0x160:
        case 0x161:
            __ACodecName="DivX Audio (WMA)";
            break;
        case 0x2000:
            __ACodecName="AC3";
            break;
        case 0x0:
        case 0x01:
            __ACodecName="Uncompressed PCM";
            break;
        }
        __M.Data[(int) MovieField.AUDIOFORMAT] = __ACodecName;

        return __M;
       

    }
    
}
